@@ -20,19 +20,15 @@ computer science today. Two models are trained simultaneously by an adversarial
 process. A generator("the artist") learns to create images that look real, while
 a discriminator ("the art critic") learns to tell real images apart from fakes.
 """
-from datetime import datetime
-import time
 import os
-import logging
-os.environ["TF_CPP_MIN_LOG_LEVEL"] = "3"  # FATAL
-logging.getLogger("tensorflow").setLevel(logging.FATAL)
+import time
+from datetime import datetime
 
 import tensorflow as tf
-
 from dnnlab.errors.dnnlab_exceptions import ModelNotCompiledError
 
 
-class CycleGAN(object):
+class CycleGAN():
     """Implements a cylce gan learning model.
 
     Attributes:
@@ -56,8 +52,14 @@ class CycleGAN(object):
         checkpoint (tf.train.Checkpoint): Groups trackable objects, saving and
             restoring them.
     """
-    def __init__(self, generator_target, generator_source,
-                 discriminator_target, discriminator_source):
+    def __init__(self,
+                 generator_target,
+                 generator_source,
+                 discriminator_target,
+                 discriminator_source,
+                 alpha_disc=0.5,
+                 alpha_identity=0.5,
+                 lambda_cycle=10):
         """Takes four keras.models that take part in the cycling gan game.
 
         Args:
@@ -70,6 +72,9 @@ class CycleGAN(object):
         self.generator_source = generator_source
         self.discriminator_target = discriminator_target
         self.discriminator_source = discriminator_source
+        self.alpha_disc = alpha_disc
+        self.alpha_identity = alpha_identity
+        self.lambda_cycle = lambda_cycle
         self.generator_target_optimizer = None
         self.generator_source_optimizer = None
         self.discriminator_target_optimizer = None
@@ -159,7 +164,7 @@ class CycleGAN(object):
         if self.generator_target_optimizer is None:
             raise ModelNotCompiledError("use CycleGAN.compile() first.")
         if mlflow:
-            import mlflow
+            import mlflow  # pylint: disable=import-outside-toplevel
 
         # Retrace workaround @function signature only tensors.
         step = tf.Variable(initial_step, name="step", dtype=tf.int64)
@@ -177,14 +182,14 @@ class CycleGAN(object):
             for source_images, target_images in zip(source_domain,
                                                     target_domain):
                 imgs = self.train_step(source_images, target_images, step,
-                                       max_outputs, file_writer)
+                                       file_writer)
 
                 # TODO Workaround. summary image & gpu usage
-                with file_writer.as_default():
+                with file_writer.as_default():  # pylint: disable=not-context-manager
                     tf.summary.image("source", imgs[0], step=step)
                     tf.summary.image("target", imgs[1], step=step)
-                    tf.summary.image("generated_source", imgs[2], step=step)
-                    tf.summary.image("generated_target", imgs[3], step=step)
+                    tf.summary.image("fake_source", imgs[2], step=step)
+                    tf.summary.image("fake_target", imgs[3], step=step)
                     tf.summary.image("cycled_source", imgs[4], step=step)
                     tf.summary.image("cycled_target", imgs[5], step=step)
 
@@ -259,8 +264,7 @@ class CycleGAN(object):
                 os.path.join(model_dir, "discriminator_source"))
 
     @tf.function
-    def train_step(self, source_images, target_images, step, max_outputs,
-                   file_writer):
+    def train_step(self, source_images, target_images, step, file_writer):
         """Decorated function (@tf.function) that creates a callable tensorflow
         graph from a python function.
         """
@@ -268,48 +272,60 @@ class CycleGAN(object):
             # Record gradients for generator and discriminator for each seperate
             # training step.
             with tf.GradientTape(persistent=True) as tape:
-                generated_target_images = self.generator_target(source_images,
-                                                                training=True)
-                cycled_source_images = self.generator_source(
-                    generated_target_images, training=True)
-
-                generated_source_images = self.generator_source(target_images,
-                                                                training=True)
-
-                cycled_target_images = self.generator_target(
-                    generated_source_images, training=True)
-
-                # same_source and same_target are used for identity loss.
-                same_source = self.generator_source(target_images,
+                # Forward cycle consistency loss.
+                # Source -> Target (single)
+                fake_target = self.generator_target(source_images,
                                                     training=True)
-                same_target = self.generator_target(source_images,
+                # Source -> Target -> Source (cycle)
+                cycled_source = self.generator_source(fake_target,
+                                                      training=True)
+
+                # Backward cycle consistency loss.
+                # Target -> Source (single)
+                fake_source = self.generator_source(target_images,
+                                                    training=True)
+                # Target -> Source -> Target (cycle)
+                cycled_target = self.generator_target(fake_source,
+                                                      training=True)
+
+                # Identity loss.
+                same_source = self.generator_source(source_images,
+                                                    training=True)
+                same_target = self.generator_target(target_images,
                                                     training=True)
 
+                # Detect real images.
                 disc_real_source = self.discriminator_source(source_images,
                                                              training=True)
                 disc_real_target = self.discriminator_target(target_images,
                                                              training=True)
 
-                disc_fake_source = self.discriminator_source(
-                    generated_source_images, training=True)
-                disc_fake_target = self.discriminator_target(
-                    generated_target_images, training=True)
+                # Detect fake images.
+                disc_fake_source = self.discriminator_source(fake_source,
+                                                             training=True)
+                disc_fake_target = self.discriminator_target(fake_target,
+                                                             training=True)
 
                 # Calculate the adversarial loss.
                 gen_target_loss = self.generator_loss(disc_fake_target)
                 gen_source_loss = self.generator_loss(disc_fake_source)
 
-                # Calculate the cycle loss
-                total_cycle_loss = self.cycle_loss(
-                    source_images, cycled_source_images) + self.cycle_loss(
-                        target_images, cycled_target_images)
+                # Calculate the cycle consistency loss.
+                total_cycle_loss = self.cycle_consistency_loss(
+                    source_images,
+                    cycled_source) + self.cycle_consistency_loss(
+                        target_images, cycled_target)
 
-                # Total gen loss = adversarial + cycle + identity
-                total_gen_target_loss = gen_target_loss + total_cycle_loss + self.identity_loss(
-                    target_images, same_target)
+                # Total generator loss = adversarial + cycle + identity
+                total_gen_target_loss = gen_target_loss + \
+                    self.lambda_cycle * total_cycle_loss + \
+                    self.identity_loss(target_images, same_target,
+                                       self.alpha_identity)
 
-                total_gen_source_loss = gen_source_loss + total_cycle_loss + self.identity_loss(
-                    source_images, same_source)
+                total_gen_source_loss = gen_source_loss + \
+                    self.lambda_cycle * total_cycle_loss + \
+                    self.identity_loss(source_images, same_source,
+                                       self.alpha_identity)
 
                 tf.summary.scalar("gen_target_loss",
                                   total_gen_target_loss,
@@ -320,9 +336,9 @@ class CycleGAN(object):
 
                 # Total discriminator loss.
                 disc_source_loss = self.discriminator_loss(
-                    disc_real_source, disc_fake_source)
+                    disc_real_source, disc_fake_source, self.alpha_disc)
                 disc_target_loss = self.discriminator_loss(
-                    disc_real_target, disc_fake_target)
+                    disc_real_target, disc_fake_target, self.alpha_disc)
                 tf.summary.scalar("disc_source_loss",
                                   disc_source_loss,
                                   step=step)
@@ -357,15 +373,14 @@ class CycleGAN(object):
                     self.discriminator_source.trainable_variables))
 
             imgs = [
-                source_images, target_images, generated_source_images,
-                generated_target_images, cycled_source_images,
-                cycled_target_images
+                source_images, target_images, fake_source, fake_target,
+                cycled_source, cycled_target
             ]
             # Piece of shit workaround.
             return imgs
 
     @staticmethod
-    def discriminator_loss(real, generated):
+    def discriminator_loss(real, fake, alpha_disc):
         """This method quantifies how well the discriminator is able to
         distinguish real images from fakes. It compares the discriminator's
         predictions on real images to an array of 1s, and the discriminator's
@@ -375,40 +390,44 @@ class CycleGAN(object):
             real_image ([type]): [description]
             generated_image ([type]): [description]
         """
+        # TODO mean least square loss for stability.
         # This method returns a helper function to compute cross entropy loss.
         cross_entropy = tf.keras.losses.BinaryCrossentropy(from_logits=True)
         # Compares real images to tensors of 1 -> if real output 1s
         real_loss = cross_entropy(tf.ones_like(real), real)
         # Compares generated images to tensors of 0 -> if generated output 0s
-        fake_loss = cross_entropy(tf.zeros_like(generated), generated)
+        fake_loss = cross_entropy(tf.zeros_like(fake), fake)
 
-        total_loss = real_loss + fake_loss
-        return total_loss * 0.5
+        total_loss = alpha_disc * (real_loss + fake_loss)
+        return total_loss
 
     @staticmethod
-    def generator_loss(generated):
+    def generator_loss(disc_decision):
         """The generator's loss quantifies how well it was able to trick the
            discriminator. Intuitively, if the generator is performing well, the
-           discriminator will classify the generated images as real (or 1).
+           discriminator will classify the fake images as real (or 1).
            Here, we will compare the discriminators decisions on the generated
            images to an array of 1s.
         """
-        # This method returns a helper function to compute cross entropy loss.
+        # TODO mean least square loss for stability.
+        # least_squared_loss = tf.keras.losses.MeanSquaredError()
         cross_entropy = tf.keras.losses.BinaryCrossentropy(from_logits=True)
-        return cross_entropy(tf.ones_like(generated), generated)
+        return cross_entropy(tf.ones_like(disc_decision), disc_decision)
 
     @staticmethod
-    def cycle_loss(real_image, cycled_image):
+    def cycle_consistency_loss(real_image, cycled_image):
         """Cycle consistency means the result should be close to the original
-        source. For example, if one translates a sentence from English to French,
+        source. For example, if one translates a sentence from English to French
         and then translates it back from French to English, then the resulting
         sentence should be the same as the original sentence.
         """
+        # Calculate mean absolute error between orig and cycled.
         return tf.reduce_mean(tf.abs(real_image - cycled_image))
 
     @staticmethod
-    def identity_loss(real_image, same_image):
+    def identity_loss(real_image, same_image, alpha_identity):
         """[summary]
         """
-        loss = tf.reduce_mean(tf.abs(real_image - same_image))
-        return loss * 0.5
+        loss = alpha_identity * (tf.reduce_mean(
+            tf.abs(real_image - same_image)))
+        return loss
