@@ -333,7 +333,7 @@ class YOLO():
                 # TODO Workaround. summary image & gpu usage.
                 with file_writer.as_default():  #pylint: disable=not-context-manager
                     tf.summary.image("label", tb_label, step=step)
-                    tf.summary.image("training", tb_prediction, step=step)
+                    tf.summary.image("prediction", tb_prediction, step=step)
 
                 file_writer.flush()
                 progbar.update(current=(step_float))
@@ -458,6 +458,8 @@ class YOLO():
 
         if model_format == "hdf5":
             self.model.save(os.path.join(model_dir, "yolo.h5"))
+        elif model_format == "SavedModel":
+            self.model.save(os.path.join(model_dir, "yolo"))
 
     @tf.function
     def train_step(self, x, y, batch_size, step, max_outputs, file_writer,
