@@ -15,4 +15,5 @@
 """DnnLab layers API"""
 from dnnlab.layers.yolo_decoder import YOLODecoder
 from dnnlab.layers.se_block import SEBlock
-from dnnlab.layers.self_attention import AttentionBlock
+from dnnlab.layers.non_local_block import NonLocalBlock
+from dnnlab.layers.morph_layer import Morph2D
